var shader_glsl = `#version 300 es
precision highp float;
in vec2 s; // position x=[-1,1] y=[-1,1] window
out vec4 r; // result
uniform float t; //time
uniform vec4 b;  //equalizer values except .w=1

// SDF PRIMITIVES AND OPERATIONS by iq

float sdBox( vec3 p, vec3 b ) { vec3 q = abs(p) - b; return length(max(q,0.0)) + min(max(q.x,max(q.y,q.z)),0.0); }

float sdPlane( vec3 p, vec3 n, float h )
{
  // n must be normalized
  return dot(p,n) + h;
}

float eskalaatio()
{
    float k = 1.-max(1.-t/10., 0.);
    
    float t0 = 23.;
    if (t > (t0-.2) && t < (t0+.2)) {
        k *= 10.*smoothstep(t0-.2, t0, t);
    }
    
    t0 = 28.5;
    if (t > (t0-.2) && t < (t0+.2)) {
        k *= 10.*smoothstep(t0-.2, t0, t);
    }
    
    t0 = 33.4;
    if (t > (t0-.2) && t < (t0+2.2)) {
        k *= 5.*smoothstep(t0-.2, t0, t);
    }
    
    t0 = 36.;
    if (t > (t0-.2) && t < (t0+.2)) {
        k *= 10.*smoothstep(t0-.2, t0, t);
    }
    
    t0 = 38.8;
    if (t > (t0-.2) && t < (t0+2.2)) {
        k *= 8.*smoothstep(t0-.2, t0, t);
    }

    return k;
}

vec3 opCheapBend2( in vec3 p )
{
    //float k = -0.1 + 0.05*sin(p.x/3. + p.z/2. + t/2.); //-.05;
    float k = -0.1 + 0.05*sin(p.x/3. + p.z/2. + t/2.); //-.05;
    k *= eskalaatio();
    //float k = calcK(p.x/3. + p.z/2. + t/2.);
    float c = cos(k*p.y);
    float s = sin(k*p.x);
    mat2  m = mat2(c,-s,s,c);
    p = vec3(m*p.xy, p.z);
    return p;
}

vec2 opU(vec2 d1, vec2 d2)
{
    return (d1.x<d2.x)?d1:d2;
}

vec3 opCheapBend3( in vec3 p )
{
    //float k = 0.02 + 0.005*sin(p.x/3. + p.z/2. + t/2. + 5.*b.z);
    float k = 0.02 + 0.005*sin(p.x/3. + p.z/2. + t/2. + 5.*b.z);
    k *= eskalaatio();
    //float k = calcK(p.x/3. + p.z/2. + t/2. + 5.*b.z);
    float c = cos(k*p.x);
    float s = sin(k*p.y);
    mat2  m = mat2(c,-s,s,c);
    return vec3(m*p.xy,p.z);
}

vec2 scene( in vec3 p)
{
    vec2 d = vec2(-.5);
    float tmp;

    vec3 s = vec3(0., 0., 20.);
    vec3 q = p;
    vec3 r = q - s*floor(q/s+0.5);
    
    
    vec3 p2 = r - vec3(0., 2., 2.);
    vec3 left = vec3(6.2, 0., 0.);

    // mirror
    p2.xy = abs(p2.xy);

    p2 = opCheapBend2(opCheapBend3(p2.xzy).xzy);

    // top & bottom
    tmp = sdPlane(p2, vec3(0., -1., 0.), 1.4*5.6);
    d = vec2(tmp, 0.);

    // left & right
    tmp = sdPlane(p2, vec3(-1., 0., 0.), 1.4*7.);
    d = opU(d, vec2(tmp, 0.));

    float len = 8.;
    /*if (t > 23. && t < 44.) {
        len = 12.;
    }*/
    tmp = sdBox(p2 - vec3(7. - t/10., 6.+b.y, 0.), vec3(4., 4., len));
    d = opU(d, vec2(tmp, 1.));

    //vec3 p3 = vec3(rotate2d(p.xy, t), p.z);
    //tmp = sdBoxFrame(p3 - vec3(0., 3., 5. + 3.*t), vec3(1.), 0.05);
    //tmp = sdBox(p3 - vec3(0., 3., 5. + 3.*t), vec3(1.));
    //d = opU(d, vec2(tmp, 1.));
 
    return d;
}

const float EPSILON = 0.001;

float dScene(vec3 p)
{
    return scene(p).x;
}

vec2 march( in vec3 ro, in vec3 rd)
{
    float t = 0.0001;
    vec2 res = vec2(-1.0);
    for (int i=0; i<512; i++) {
        vec3 p = ro + t*rd;
        vec2 d = scene(p);
        if (d.x < 0.01) {
            res = vec2(t, d.y);
            break;
        }
        if (t > 100.0) {
            break;
        }
        
        t += d.x;
    }
    return res;
}

vec3 render( in vec3 ro, in vec3 rd)
{
    vec3 fog = vec3(b.rbg)/3.;
    vec3 col = fog;
    vec2 res = march(ro, rd);

    if (res.x > -0.5) {
        //vec3 N = sceneNormal(ro + res.x*rd);
        
        float mat = res.y;
        if (mat < 0.5) {
            col = b.rbg;
        } else {
            col = vec3(.0);
        }
        
        col = mix(col, fog, pow(max(min(res.x/20.0-.1, 1.0), 0.), 2.));
    }
    return col;
}

void main()
{
    vec2 uv = s*vec2(16./9.,1.);
    vec3 rd = normalize(vec3(uv.x, uv.y, 1));
    vec3 ro = vec3(0., 3. + .5*sin(t), -1.+3.*t + b.z);

    vec3 col = render(ro, rd);
    col = pow( col, vec3(0.4545) );

    r = vec4(col, 1.0);
}`;